
class ClientCompliance:
    # 0 - zero errors, all ok
    # 1 - unknown, not enough information
    # 2 - warning, something isn't quite right
    # 3 - critical, something definately is wrong
    STATE_OK = 0
    STATE_UNKNOWN = 1
    STATE_WARNING = 2
    STATE_ERROR = 3

    def __init__(self, total_disk, free_disk, last_checkin_days, outstanding_critical_patches = -1, outstanding_standard_patches = -1):
        self.total_disk = total_disk
        self.free_disk = free_disk
        self.last_checkin_days = last_checkin_days
        self.outstanding_critical_patches = outstanding_critical_patches
        self.outstanding_standard_patches = outstanding_standard_patches

    def get_patch_compliance(self):
        if self.outstanding_critical_patches == 0 and self.outstanding_standard_patches == 0:
            return ClientCompliance.STATE_OK

        if self.outstanding_critical_patches > 0:
            return ClientCompliance.STATE_ERROR

        if self.outstanding_standard_patches > 0:
            return ClientCompliance.STATE_WARNING

        return ClientCompliance.STATE_UNKNOWN

    def get_checkin_compliance(self):
        if self.last_checkin_days is None:
            return ClientCompliance.STATE_UNKNOWN
        if self.last_checkin_days < 7:
            return ClientCompliance.STATE_OK
        if self.last_checkin_days < 14:
            return ClientCompliance.STATE_WARNING
        return ClientCompliance.STATE_ERROR

    def get_disk_compliance(self):
        if self.free_disk is None or self.total_disk is None:
            return ClientCompliance.STATE_UNKNOWN
            
        # < 20% left is warning
        # < 5% left is critical, or less than 5g
        space_left_pcnt = (float(self.free_disk) / float(self.total_disk)) * 100.0
        
        space_compliance = ClientCompliance.STATE_UNKNOWN
        if space_left_pcnt >= 20:
            space_compliance = ClientCompliance.STATE_OK
        elif space_left_pcnt < 5:
            space_compliance = ClientCompliance.STATE_ERROR
        else:
            space_compliance = ClientCompliance.STATE_WARNING # its just less than 20

        return space_compliance

    def get_compliance_state(self):
        checkin = self.get_checkin_compliance()        
        disk = self.get_disk_compliance()
        patch = self.get_patch_compliance()
        return max(checkin, disk, patch)



MAIN_GROUP_ID = 5
MAIN_GROUP_NAME = "I am your Group"
MAIN_GROUP_PARENT_ID = 433

INVALID_GROUP = 55

test_win_app_generic_should_not_load = '''
{
    "criteria": {
        "expressions": [
            {
                "column": "type",
                "component": "OperatingSystem",
                "operator": "is",
                "qualifier": "WIN"
            }
        ],
        "logic": "all"
    },
    "fields": [
        {
            "column": "product_id",
            "component": "Application"
        },
        {
            "column": "name",
            "component": "Application"
        },
        {
            "column": "version",
            "component": "Application"
        },
        {
            "column": "size",
            "component": "Application"
        },
        {
            "column": "install_size",
            "component": "Application"
        },
        {
            "column": "is_validated",
            "component": "Application"
        },
        {
            "column": "device_name",
            "component": "Client"
        }
    ],
    "main_component": "Application",
    "name": "generic query",
    "id": 100,
    "group": %d
}
''' % MAIN_GROUP_ID

test_query_app_adobe_acrobat_reader_win = '''{
    "criteria": {
        "expressions": [
            {
                "column": "name",
                "component": "Application",
                "operator": "is",
                "qualifier": "Adobe Acrobat Reader DC"
            },
            {
                "column": "type",
                "component": "OperatingSystem",
                "operator": "is",
                "qualifier": "WIN"
            }
        ],
        "logic": "all"
    },
    "fields": [
        {
            "column": "product_id",
            "component": "Application"
        },
        {
            "column": "name",
            "component": "Application"
        },
        {
            "column": "version",
            "component": "Application"
        },
        {
            "column": "size",
            "component": "Application"
        },
        {
            "column": "install_size",
            "component": "Application"
        },
        {
            "column": "is_validated",
            "component": "Application"
        },
        {
            "column": "filewave_client_name",
            "component": "Client"
        },
        {
            "column": "device_id",
            "component": "Client"
        }
    ],
    "main_component": "Application",
    "id": 101,
    "group": %d,
    "name": "Adobe Acrobat Reader Win"
}''' % MAIN_GROUP_ID

test_query_app_zoom_win = '''{
    "criteria": {
        "expressions": [
            {
                "column": "name",
                "component": "Application",
                "operator": "is",
                "qualifier": "Zoom"
            },
            {
                "column": "type",
                "component": "OperatingSystem",
                "operator": "is",
                "qualifier": "WIN"
            },
            {
                "column": "version",
                "component": "Application",
                "operator": "contains",
                "qualifier": "."
            }
        ],
        "logic": "all"
    },
    "fields": [
        {
            "column": "product_id",
            "component": "Application"
        },
        {
            "column": "name",
            "component": "Application"
        },
        {
            "column": "version",
            "component": "Application"
        },
        {
            "column": "size",
            "component": "Application"
        },
        {
            "column": "install_size",
            "component": "Application"
        },
        {
            "column": "is_validated",
            "component": "Application"
        },
        {
            "column": "device_id",
            "component": "Client"
        }
    ],
    "main_component": "Application",
    "group": %d,
    "id": 105,
    "name": "Zoom Win"
}''' % MAIN_GROUP_ID

test_query_app_wrong_group_zoom_win = '''{
    "criteria": {
        "expressions": [
            {
                "column": "name",
                "component": "Application",
                "operator": "is",
                "qualifier": "Zoom"
            },
            {
                "column": "type",
                "component": "OperatingSystem",
                "operator": "is",
                "qualifier": "WIN"
            },
            {
                "column": "version",
                "component": "Application",
                "operator": "contains",
                "qualifier": "."
            }
        ],
        "logic": "all"
    },
    "fields": [
        {
            "column": "product_id",
            "component": "Application"
        },
        {
            "column": "name",
            "component": "Application"
        },
        {
            "column": "version",
            "component": "Application"
        },
        {
            "column": "size",
            "component": "Application"
        },
        {
            "column": "install_size",
            "component": "Application"
        },
        {
            "column": "is_validated",
            "component": "Application"
        },
        {
            "column": "device_id",
            "component": "Client"
        }
    ],
    "group": %d,
    "main_component": "Application",
    "name": "Zoom Win wrong group",
    "id": 120,
    "favorite": false
}''' % INVALID_GROUP


from extra_metrics.main import serve_and_process
from extra_metrics.config import ExtraMetricsConfiguration, read_config_helper
from extra_metrics.logs import logger

import json
import os
import click
import shutil
import pkg_resources
import subprocess
import re
import sys

class ValidationExceptionCannotParseFileWaveVersion(Exception):
    pass

class ValidationExceptionWrongFileWaveVersion(Exception):
    pass

@click.group()
def cli():
    click.echo("FileWave Extra Metrics configuration.")


@cli.command('validate', help="Validates the configuration assuming you are running this on the FileWave Server")
@click.option('-c', '--config-path', 'config_path', default=ExtraMetricsConfiguration.DEFAULT_CFG_FILE_LOCATION, help='the full path to the configuration file')
@click.option('-a', '--api-key', 'api_key', default=None, help='the FileWave API Key with appropriate rights to create groups/queries')
@click.option('-e', '--external-dns-name', 'external_dns_name', default=None, help='the externally visible DNS name for the filewave server')
@click.option('-v', '--validate', 'validate', default=True, help='if config is present, validate some basic requirements by calling the FW server using the API Key')
def install_into_environment(config_path, api_key, external_dns_name, validate):
    cfg = ExtraMetricsConfiguration()
    
    # TODO: get the API key from the bearer_token_file - abort if not there.  assumes we are running localhost

    dirname = os.path.dirname(config_path)
    if not os.path.exists(dirname):
        logger.error(f"The directory for the configuration file does not exist: {dirname}")
        return

    if not os.path.exists(config_path) and os.path.isfile(config_path):
        if not os.access(config_path, os.W_OK):
            logger.error(f"The configuration file cannot be written to {config_path} - does this user have access?")
            return

    try:
        read_config_helper(cfg)
    except FileNotFoundError:
        if api_key is None or external_dns_name is None:
            logger.error("When there is no configuration file you must specify an API key and external DNS name, which will then be stored in the config file")
            return

    assert cfg.section is not None

    if api_key is not None:
        cfg.set_fw_api_key(api_key)
    if external_dns_name is not None:
        cfg.set_fw_api_server(external_dns_name)

    try:
        with open(config_path, 'w+') as f:
            cfg.write_configuration(f)
            logger.info(f"saved configuration to file: {config_path}")
    except Exception as e:
        logger.error("Unable to write the configuration file - normally this command requires sudo/root privs, did you use sudo?")
        logger.error(e)
        return

    try:    
        provision_dashboards_into_grafana(cfg.get_fw_api_server())
        provision_prometheus_scrape_configuration()
        provision_supervisord_runtime()
    except Exception as e:
        logger.error("Error during provisioning of prometheus/grafana, are you using sudo?")
        logger.error(e)
        return

    logger.info("")
    logger.info("Configuration Summary")
    logger.info("=====================")
    logger.info(f"API Key: {cfg.get_fw_api_key()}")
    logger.info(f"External DNS: {cfg.get_fw_api_server()}")

    validate_runtime_requirements(cfg)


def provision_supervisord_runtime():
    supervisord_dir = os.path.join("/usr/local/etc/filewave/supervisor/", "extras")
    if not os.path.exists(supervisord_dir):
        os.makedirs(supervisord_dir)

    # lets go hunting, I want to find executable 'extra_metrics_run' - which is installed by setuptool,
    # but I don't know where the whole thing is being run from... so I'm looking for bin/extra_metrics_run 
    # starting with the sys.executable directory
    exec_path = os.path.dirname(sys.executable)
    full_extra_metrics_run_path = "extra-metrics-run"
    for f in os.listdir(exec_path):
        if f == "extra-metrics-run":
            full_extra_metrics_run_path = os.path.join(exec_path, full_extra_metrics_run_path)

    data = pkg_resources.resource_string("extra_metrics.cfg", "extra_metrics_supervisord.conf").decode('utf-8')
    provisioning_file = os.path.join(supervisord_dir, 'extra_metrics_supervisord.conf')
    with open(provisioning_file, "w+") as f:
        new_data = data.replace(r'${EXTRA_METRICS_RUN}', full_extra_metrics_run_path)
        f.write(new_data)


def provision_dashboards_into_grafana(fw_server_dns_name):
    # if the expected dashboards DO NOT exist in the right path, moan about this and go ahead and copy them..
    grafana_dashboard_deployment_dir = os.path.join("/usr/local/etc/filewave/grafana/provisioning", "dashboards")
    if not os.path.exists(grafana_dashboard_deployment_dir):
        logger.error(f"The Grafana dashboard deployment directory ({grafana_dashboard_deployment_dir}) does not exist; is this version 14+ of FileWave?")
        return

    # check each file is there... overwrite regardless (helps on upgrade I suppose)
    for dashboard_file in pkg_resources.resource_listdir("extra_metrics", "dashboards"):
        if dashboard_file.endswith(".json"):
            data = pkg_resources.resource_string("extra_metrics.dashboards", dashboard_file).decode('utf-8')
            provisioning_file = os.path.join(grafana_dashboard_deployment_dir, dashboard_file)
            with open(provisioning_file, 'w+') as f:
                # load up the dashboard and replace the ${VAR_SERVER} with our config value of the external DNS
                new_data = data.replace('${VAR_SERVER}', fw_server_dns_name)
                f.write(new_data)
                logger.info(f"wrote dashboard file: {provisioning_file}")


def provision_prometheus_scrape_configuration():
    prometheus_dir = os.path.join("/usr/local/etc/filewave/prometheus/conf.d/jobs", "http")
    if not os.path.exists(prometheus_dir):
        logger.error(f"The Prometheus directory ({prometheus_dir}) does not exist; is this version 14+ of FileWave?")
        return

    for yaml_file in pkg_resources.resource_listdir("extra_metrics", "cfg"):
        if yaml_file.endswith(".yml"):
            data = pkg_resources.resource_string("extra_metrics.cfg", yaml_file)
            provisioning_file = os.path.join(prometheus_dir, yaml_file)
            with open(provisioning_file, 'wb') as f:
                f.write(data)
            shutil.chown(provisioning_file, user="apache", group="apache")


def get_current_fw_version():
    proc = subprocess.Popen(["/usr/local/bin/fwcontrol", "server", "version"], stdout=subprocess.PIPE)
    return proc.communicate()[0]


def validate_current_fw_version():
    current_ver = get_current_fw_version()
    exp = re.compile(r'fwxserver (\d+).(\d+).(\d+)', re.IGNORECASE)
    match_result = re.search(exp, current_ver)

    major, minor, patch = 0, 0, 0
    if match_result is not None:
        major = int(match_result.group(1))
        minor = int(match_result.group(2))
        patch = int(match_result.group(3))
        logger.info(f"detected FileWave instance running version: {major}.{minor}.{patch}")
    else:
        raise ValidationExceptionCannotParseFileWaveVersion("Failed to detection which version of FileWave is running")
    
    if major < 14:
        err = f"You must be running FileWave version 14 or above - I detected a FileWave server version: {major}.{minor}.{patch}"
        logger.error(err)
        raise ValidationExceptionWrongFileWaveVersion(err)


def validate_runtime_requirements(cfg):
    # TODO: does the API key have the appropriate rights? 
    # TODO: can I upgrade the grafana pie chart automatically? 
    validate_current_fw_version()
    




import logging
import sys

logger = logging.getLogger('extra-metrics')

def init_logging():
    ch = logging.StreamHandler(sys.stdout)
    ch.setLevel(logging.INFO)
    formatter = logging.Formatter('[%(asctime)s] [%(name)s] [%(levelname)s] %(message)s')
    ch.setFormatter(formatter)
    logger.addHandler(ch)
    logger.setLevel(logging.INFO)



from prometheus_client import Gauge
import pandas as pd
import datetime
from extra_metrics.compliance import ClientCompliance
from extra_metrics.logs import logger

device_checkin_days = Gauge('extra_metrics_devices_by_checkin_days',
                            'various interesting stats on a per device basis, days since checked, compliance status',
                            ["days", ])

device_client_compliance = Gauge('extra_metrics_per_device_compliance',
                           'provides compliance summary on a per device basis, the value here is the device model number',
                           ["device_name", "compliant"])

device_client_version = Gauge('extra_metrics_per_device_client_version',
                           'number of devices rolled up by client version',
                           ["fw_client_version"])

device_client_platform = Gauge('extra_metrics_per_device_platform',
                           'number of devices rolled up by platform',
                           ["platform"])

device_client_tracked = Gauge('extra_metrics_per_device_tracked',
                           'number of devices being tracked',
                           ["tracked"])

device_client_locked = Gauge('extra_metrics_per_device_locked',
                           'number of devices locked',
                           ["locked"])

class PerDeviceStatus:
    def __init__(self, fw_query):
        self.fw_query = fw_query
        self.state_by_patch = {}

    def _rollup_by_single_column_count_client_filewave_id(self, df, column_name):
        return df.groupby([column_name], as_index=False)["Client_filewave_id"].count()

    def _set_metric_pair(self, metric, item):
        label_value = item[0]
        total_count = item[1]
        metric.labels(label_value).set(total_count)
        return (label_value, total_count)

    def collect_client_data(self):
        Client_device_name = 0
        Client_free_disk_space = 2
        Client_last_check_in = 17
        DesktopClient_filewave_model_number = 18
        Client_total_disk_space = 24

        r = self.fw_query.get_client_info()
        j = r.json()

        try:
            assert j["fields"]
            assert j["fields"][Client_device_name] == "Client_device_name", "field 0 is expected to be the Client's name"
            assert j["fields"][Client_last_check_in] == "Client_last_check_in", "field 17 is expected to be the Client's last check in date/time"

            buckets = [0, 0, 0, 0]
            now = datetime.datetime.now()

            df = pd.DataFrame(j["values"], columns=j["fields"])

            # devices by client_version
            for item in self._rollup_by_single_column_count_client_filewave_id(df, "DesktopClient_filewave_client_version").to_numpy():
                version = item[0]
                total_count = item[1]
                if version is None:
                    version = "Not Reported"
                device_client_version.labels(version).set(total_count)
                logger.info(f"device client version: {version}, {total_count}")

            # roll up devices per platform
            for item in self._rollup_by_single_column_count_client_filewave_id(df, "OperatingSystem_name").to_numpy():
                (a, b) = self._set_metric_pair(device_client_platform, item)
                logger.info(f"device platform: {a}, {b}")

            # roll up devices by 'tracking enabled' or not
            for item in self._rollup_by_single_column_count_client_filewave_id(df, "Client_is_tracking_enabled").to_numpy():
                (a, b) = self._set_metric_pair(device_client_tracked, item)
                logger.info(f"device by tracking: {a}, {b}")

            # and by locked state
            for item in self._rollup_by_single_column_count_client_filewave_id(df, "Client_filewave_client_locked").to_numpy():
                (a, b) = self._set_metric_pair(device_client_locked, item)
                logger.info(f"device by locked: {a}, {b}")

            # a bit of logic here, so rollup isn't via pandas...

            for v in j["values"]:
                # if there is no last check in date, we want to assume it's NEVER checked in
                checkin_days = 99
                if v[Client_last_check_in] is not None:
                    checkin_date = datetime.datetime.strptime(
                        v[Client_last_check_in], '%Y-%m-%dT%H:%M:%S.%fZ')
                    delta = now - checkin_date
                    checkin_days = delta.days

                comp_check = ClientCompliance(
                    v[Client_total_disk_space],
                    v[Client_free_disk_space],
                    checkin_days
                )

                # TODO: when rolling this up, if we have another entry that is non-null in any of the columns
                # and this row IS null; drop this row, e.g. nuke duplicates - this work is pending tests in a larger environment.
                device_client_compliance.labels(
                    v[Client_device_name],
                    comp_check.get_compliance_state() 
                ).set(v[DesktopClient_filewave_model_number] if v[DesktopClient_filewave_model_number] is not None else 0)

                if(checkin_days <= 1):
                    buckets[0] += 1
                elif checkin_days < 7:
                    buckets[1] += 1
                elif checkin_days < 30:
                    buckets[2] += 1
                else:
                    buckets[3] += 1

            # TODO: languages / translation?

            device_checkin_days.labels('Less than 1').set(buckets[0])
            device_checkin_days.labels('Less than 7').set(buckets[1])
            device_checkin_days.labels('Less than 30').set(buckets[2])
            device_checkin_days.labels('More than 30').set(buckets[3])

        except AssertionError as e1:
            print("The validation/assertions failed: %s" % (e1,))



from prometheus_client import start_http_server, Gauge

import logging
import timeloop
import random
import time
import sys
import os, json
import datetime
import threading
import pandas as pd 

from extra_metrics.application import ApplicationQueryManager
from extra_metrics.softwarepatches import SoftwarePatchStatus
from extra_metrics.devices import PerDeviceStatus
from extra_metrics.fwrest import FWRestQuery
from extra_metrics.config import ExtraMetricsConfiguration, read_config_helper
from extra_metrics.logs import logger, init_logging

init_logging()

# TODO: inject the supervisord configuration 

# TODO: DEVICE HEALTH show the reasons/state of health of a device in custom fields so it can be reported on

# TODO: there should be a top10 popular/requested updates table/panel. 

# TODO: link to devices affected in the dashboard is wrong; we can fix that!

# TODO: consider alerts for devices entering a non-healthy state for the first time today

# TODO: consider how I might achieve a canibalization test for devices? 

# TODO: write documentation on arch and reasoning... perhaps using the Wiki in GitHub.

tl = timeloop.Timeloop()

app_qm = None
software_patches = None
per_device = None

pd.set_option('display.precision', 3)
pd.set_option('display.expand_frame_repr', False)


@tl.job(interval=datetime.timedelta(seconds=30))
def validate_and_collect_data():
    app_qm.validate_query_definitions()
    app_qm.collect_application_query_results()

    per_device.collect_client_data()

    software_patches.collect_patch_data_status()
    software_patches.collect_patch_data_per_device()


def serve_and_process():
    cfg = ExtraMetricsConfiguration()
    read_config_helper(cfg)

    # TODO: can I use this?  get the hostname via curl -X GET "https://fwsrv.cluster8.tech/api/config/app"
    fw_query = FWRestQuery(hostname=cfg.get_fw_api_server(), api_key=cfg.get_fw_api_key())

    global app_qm, software_patches, per_device

    app_qm = ApplicationQueryManager(fw_query)
    software_patches = SoftwarePatchStatus(fw_query)
    per_device = PerDeviceStatus(fw_query)

    # be a web server... go on...
    start_http_server(8000)

    # collect the first chunk of info from our server
    validate_and_collect_data()

    # just sit here being a web server...
    tl.start(block=True)





from prometheus_client import Gauge

import traceback
import pkg_resources
import pandas as pd
import os, json, sys

from .fwrest import FWRestQuery, http_request_time_taken
from .logs import logger


app_version_count = Gauge('extra_metrics_application_version',
    'a summary of how many devices are using a particular app & version',
    ["application_name", "application_version", "query_id"])


"""
Application Usage - Rollup

The idea is to run a FW query, use Pandas to do a [multi] column rollup and count the number of instances 
per version of the application.

Once the counts have been made; they are stored and would be ready to pull out and put into a metric, that 
isn't part of this class though.
"""
class ApplicationUsageRollup:
    def __init__(self, query_id, rollup_column_names, op_column_name, op="count"):
        self.query_id = query_id
        self.rollup_column_names = rollup_column_names if isinstance(rollup_column_names, list) else [rollup_column_names]
        self.op_column_name = op_column_name
        self.op = op
        self.result_df = None

    def exec(self, fw_rest_api):
        r = fw_rest_api.get_results_for_query_id(self.query_id)
        j = r.json()
        df = pd.DataFrame(j["values"], columns=j["fields"])
        # run the group-by and count operation
        self.result_df = df.groupby(self.rollup_column_names, as_index=False)[self.op_column_name].count()

    def results(self):
        return self.result_df.to_numpy()

"""
Responsible for managing the refresh of queries for apps from the FW server, associating that 
with the rolled up results (via ApplicationUsageRollup)
"""
class ApplicationQueryManager:
    def __init__(self, fw_query):
        self.fw_query = fw_query
        self.app_queries = {}

    def is_query_valid(self, q_id):
        app_name = False
        app_version = False
        client_id = False

        # fetch the query def
        json_query = self.fw_query.get_definition_for_query_id_j(q_id)
        if json_query is None:
            return False

        # requires id/name too
        if "name" not in json_query:
            return False

        if "id" not in json_query:
            return False

        # we are expecting Application_name, Application_version and Client_filewave_id
        fields_array = json_query["fields"]
        for item in fields_array:
            if item["column"] == "name" and item["component"] == "Application":
                app_name = True
            if item["column"] == "version" and item["component"] == "Application":
                app_version = True
            if item["column"] == "device_id" and item["component"] == "Client":
                client_id = True

        return app_name and app_version and client_id

    def create_default_queries_in_group(self, group_id):
        for query_file in pkg_resources.resource_listdir("extra_metrics", "app_queries"):
            if query_file.endswith(".json"):
                f = pkg_resources.resource_string("extra_metrics.app_queries", query_file)
                json_data = json.loads(f)
                json_data["group"] = group_id
                self.fw_query.create_inventory_query(json.dumps(json_data))

    def retrieve_all_queries_in_group(self, group_id):
        # in all cases - read all queries in this group and find their IDs, store in-mem for redirection 
        all_queries = self.fw_query.get_all_inventory_queries()
        
        self.app_queries = {}
        for q in all_queries:
            q_group = q["group"]
            q_id = q["id"]
            if q_group == group_id and self.is_query_valid(q_id):
                self.app_queries[q_id] = q
                logger.info(f"refreshed app query {q_id}/{q['name']}")

    def validate_query_definitions(self):
        # ensure there is a top level query group we can place queries into...
        query_group, was_created = self.fw_query.ensure_inventory_query_group_exists("Extra Metrics Queries - Apps")
        group_id = int(query_group['id'])

        # if we created this group; then pre-pop the queries with what we have in code, otherwise
        # assume the queries inside this group are definitive
        if was_created:
            self.create_default_queries_in_group(group_id)

        self.retrieve_all_queries_in_group(group_id)

    def collect_application_query_results(self):
        for q_id, q in self.app_queries.items():
            r = ApplicationUsageRollup(q_id, ["Application_name", "Application_version"], "Client_device_id")

            try:
                label_name = f"app_query_{q['name']}"
                with http_request_time_taken.labels(label_name).time():
                    r.exec(self.fw_query)
                
                # and of course, throw this into the metric we are exposing.
                for result in r.results():
                    name = result[0]
                    version = result[1]
                    total = int(result[2])
                    logger.info(f"app query result for {name}, {version}, query_id: {r.query_id} = {total}")
                    app_version_count.labels(name, version, r.query_id).set(total)                
            except Exception as e:
                logger.error(f"failed to do app query rollup on query id {q_id}, {e}")
                traceback.print_exc(file=sys.stdout)


from extra_metrics.fwrest import FWRestQuery

import json
import sys
import os
import io
import pandas as pd
import unittest
import unittest.mock as mock

from prometheus_client import REGISTRY

from extra_metrics.application import ApplicationQueryManager, ApplicationUsageRollup, app_version_count
from extra_metrics.compliance import ClientCompliance
from extra_metrics.config import ExtraMetricsConfiguration
from .test_queries import *

from extra_metrics.scripts import get_current_fw_version, provision_supervisord_runtime, \
    validate_current_fw_version, ValidationExceptionCannotParseFileWaveVersion, ValidationExceptionWrongFileWaveVersion

pd.set_option('display.precision', 3)
pd.set_option('display.expand_frame_repr', False)

class FakeRequest:
    def __init__(self, string_data):
        self.data = string_data

    def json(self):
        return json.loads(self.data)


class FakeQueryInteface:
    TEST_QUERY_APPS = 1
    TEST_QUERY_DATA_INVALID = 99
    TEST_QUERY_DATA_ALMOST_VALID1 = 98
    TEST_QUERY_DATA_ALMOST_VALID2 = 97

    def __init__(self, create_inventory_callback=None):
        self.create_inventory_callback = create_inventory_callback
        super()

    def create_inventory_query(self, json_obj):
        if self.create_inventory_callback:  # pragma: no branch
            self.create_inventory_callback(json_obj)

    def ensure_inventory_query_group_exists(self, name_of_query):
        assert name_of_query is not None
        my_data = '''{
            "name": "%s",
            "parent": %d,
            "id": %d
        }''' % (MAIN_GROUP_NAME, MAIN_GROUP_PARENT_ID, MAIN_GROUP_ID)
        return json.loads(my_data), False

    def get_all_inventory_queries(self):
        # returns some good, some bad... some in our group, some not.
        text_queries = [test_query_app_adobe_acrobat_reader_win,
                        test_win_app_generic_should_not_load,
                        test_query_app_wrong_group_zoom_win,
                        test_query_app_zoom_win]

        return [json.loads(t) for t in text_queries]

    def get_definition_for_query_id_j(self, q_id):
        if q_id == FakeQueryInteface.TEST_QUERY_DATA_INVALID:
            return None

        if q_id == FakeQueryInteface.TEST_QUERY_DATA_ALMOST_VALID1:
            badly_formed_query = '''{
                "favorite": true,
                "fields": [
                    {
                        "column": "filewave_id",
                        "component": "Client"
                    },
                    {
                        "column": "filewave_client_name",
                        "component": "Client"
                    },
                    {
                        "column": "version",
                        "component": "Update"
                    }
                ],
                "main_component": "Update"
                }'''
            return json.loads(badly_formed_query)

        if q_id == FakeQueryInteface.TEST_QUERY_DATA_ALMOST_VALID2:
            badly_formed_query = '''{
                "favorite": true,
                "fields": [
                    {
                        "column": "filewave_id",
                        "component": "Client"
                    },
                    {
                        "column": "filewave_client_name",
                        "component": "Client"
                    },
                    {
                        "column": "version",
                        "component": "Update"
                    }
                ],
                "name": "this has a name",
                "main_component": "Update"
                }'''
            return json.loads(badly_formed_query)

        if q_id == 66:
            badly_formed_query = '''{
                "favorite": true,
                "fields": [
                    {
                        "column": "filewave_id",
                        "component": "Client"
                    },
                    {
                        "column": "filewave_client_name",
                        "component": "Client"
                    },
                    {
                        "column": "version",
                        "component": "Update"
                    }
                ],
                "main_component": "Update",
                "id": 66,
                "name": "extra metrics - software patch"
                }'''
            return json.loads(badly_formed_query)

        if q_id == 55:
            well_formed_query = '''{
                "favorite": true,
                "fields": [
                    {
                        "column": "device_id",
                        "component": "Client"
                    },
                    {
                        "column": "name",
                        "component": "Application"
                    },
                    {
                        "column": "version",
                        "component": "Application"
                    }
                ],
                "main_component": "Update",
                "name": "extra metrics - software patch",
                "id": 55
            }'''
            return json.loads(well_formed_query)

        all_queries = self.get_all_inventory_queries()
        for a_query in all_queries:
            if a_query["id"] == q_id:
                return a_query
        return None

    def get_results_for_query_id(self, query_id):
        if query_id == 105:
            return FakeRequest('''{
                "offset": 0,
                "fields": [
                    "Application_product_id",
                    "Application_name",
                    "Application_version",
                    "Application_size",
                    "Application_install_size",
                    "Application_is_validated",
                    "Client_filewave_client_name",
                    "Client_device_id"
                ],
                "values": [
                    [
                        "{abcdefg}",
                        "Zoom",
                        "12.1",
                        null,
                        6459167232,
                        null,
                        "hades",
                        "4d9fe4590232f39b783d95adf8b12bb24e4f7139"
                    ],
                    [
                        "{AC76BA86-7AD7-1033-7B44-AC0F074E4100}",
                        "Zoom",
                        "20",
                        null,
                        339167232,
                        null,
                        "hades",
                        "4d9fe4590232f39b783d95adf8b12bb24e4f7139"
                    ],
                    [
                        "{AC76BA86-7AD7-1033-7B44-AC0F074E4100}",
                        "Zoom App",
                        "20.1",
                        null,
                        339167232,
                        null,
                        "hades",
                        "4d9fe4590232f39b783d95bdf8b12bb24e4f7139"
                    ]

                ],
                "filter_results": 3,
                "total_results": 3,
                "version": 1
            }''')

        if query_id == FakeQueryInteface.TEST_QUERY_APPS or query_id == 101:
            return FakeRequest('''{
                "offset": 0,
                "fields": [
                    "Application_product_id",
                    "Application_name",
                    "Application_version",
                    "Application_size",
                    "Application_install_size",
                    "Application_is_validated",
                    "Client_filewave_client_name",
                    "Client_device_id"
                ],
                "values": [
                    [
                        "{abcdefg}",
                        "Adobe Acrobat Reader DC",
                        "12.1",
                        null,
                        6459167232,
                        null,
                        "hades",
                        "4d9fe4590232f39b783d95adf8b12bb24e4f7139"
                    ],
                    [
                        "{AC76BA86-7AD7-1033-7B44-AC0F074E4100}",
                        "Adobe Acrobat Reader DC",
                        "20.009.20067",
                        null,
                        339167232,
                        null,
                        "hades",
                        "4d9fe4590232f39b783d95adf8b12bb24e4f7139"
                    ],
                    [
                        "{AC76BA86-7AD7-1033-7B44-AC0F074E4100}",
                        "Adobe Acrobat Reader DC",
                        "20.009.20067",
                        null,
                        339167232,
                        null,
                        "hades2",
                        "4d9fe4590232f39b783d95bdf8b12bb24e4f7139"
                    ]

                ],
                "filter_results": 3,
                "total_results": 3,
                "version": 1
            }''')


fw_query = FakeQueryInteface()


class ExtraMetricsQueryTestCase(unittest.TestCase):
    def test_query_init(self):
        fq = FWRestQuery("a", "b")
        self.assertEqual(fq.hostname, "a")
        self.assertEqual(fq.api_key, "b")

    def test_query_inventory_stub(self):
        fq = FWRestQuery("a", "b")
        self.assertEqual("https://a:20445/inv/api/v1/def",
                         fq._fw_run_inv_query("def"))
        self.assertEqual("https://a/api/xyz", fq._fw_run_web_query("xyz"))


class ExtraMetricsTestCase(unittest.TestCase):
    def test_app_manager_rolls_queries_and_populates_metrics(self):
        app_mgr = ApplicationQueryManager(fw_query)
        app_mgr.validate_query_definitions()
        app_mgr.collect_application_query_results()

        after = REGISTRY.get_sample_value('extra_metrics_application_version',
                                          labels={"application_name": "Adobe Acrobat Reader DC", "application_version": "12.1", "query_id": "101"})
        self.assertEqual(1, after)

        after = REGISTRY.get_sample_value('extra_metrics_application_version',
                                          labels={"application_name": "Adobe Acrobat Reader DC", "application_version": "20.009.20067", "query_id": "101"})
        self.assertEqual(2, after)

        after = REGISTRY.get_sample_value('extra_metrics_application_version',
                                          labels={"application_name": "Zoom App", "application_version": "20.1", "query_id": "105"})
        self.assertEqual(1, after)

    def test_app_manager_query_validation(self):
        app_mgr = ApplicationQueryManager(fw_query)
        self.assertEqual(len(app_mgr.app_queries), 0)
        app_mgr.validate_query_definitions()
        self.assertEqual(len(app_mgr.app_queries), 2)

    def test_app_manager_invalid_query_cannot_validate(self):
        app_mgr = ApplicationQueryManager(fw_query)
        self.assertFalse(app_mgr.is_query_valid(FakeQueryInteface.TEST_QUERY_DATA_INVALID))
        self.assertFalse(app_mgr.is_query_valid(FakeQueryInteface.TEST_QUERY_DATA_ALMOST_VALID1))
        self.assertFalse(app_mgr.is_query_valid(FakeQueryInteface.TEST_QUERY_DATA_ALMOST_VALID2))

    def test_app_manager_doesnt_load_crap_queries(self):
        app_mgr = ApplicationQueryManager(fw_query)
        app_mgr.retrieve_all_queries_in_group(MAIN_GROUP_ID)
        self.assertEqual(len(app_mgr.app_queries), 2)
        self.assertTrue(120 not in app_mgr.app_queries.keys())
        self.assertTrue(101 in app_mgr.app_queries.keys())
        self.assertTrue(105 in app_mgr.app_queries.keys())

    def test_app_mgr_accepts_valid_inventory_query(self):
        app_mgr = ApplicationQueryManager(fw_query)
        result = app_mgr.is_query_valid(55)
        self.assertTrue(result)

    def test_app_mgr_rejects_invalid_inventory_query(self):
        app_mgr = ApplicationQueryManager(fw_query)
        result = app_mgr.is_query_valid(66)
        self.assertFalse(result)

    def test_app_rollup(self):
        thing = ApplicationUsageRollup(FakeQueryInteface.TEST_QUERY_APPS, [
                                       "Application_name", "Application_version"], "Client_device_id")
        thing.exec(fw_query)

        # examine the rollup results to prove this worked
        r = thing.results()
        self.assertEqual(len(r), 2)

        item = r[0]
        item_name = item[0]
        item_version = item[1]
        item_count = item[2]

        self.assertEqual(item_name, "Adobe Acrobat Reader DC")
        self.assertEqual(item_version, "12.1")
        self.assertEqual(item_count, 1)

        item = r[1]
        item_name = item[0]
        item_version = item[1]
        item_count = item[2]

        self.assertEqual(item_name, "Adobe Acrobat Reader DC")
        self.assertEqual(item_version, "20.009.20067")
        self.assertEqual(item_count, 2)

    def test_app_mgr_can_create_queries(self):
        loaded_data = []

        def test_query_load(json_data):
            nonlocal loaded_data
            loaded_data.append(json_data)

        my_query = FakeQueryInteface(test_query_load)
        app_mgr = ApplicationQueryManager(my_query)
        app_mgr.create_default_queries_in_group(MAIN_GROUP_ID)
        self.assertEqual(9, len(loaded_data))
        # for r in loaded_data:
        #     print(r)


class TestClientComplianceCase(unittest.TestCase):
    def test_compliance_with_no_data(self):
        c = ClientCompliance(None, None, None)
        self.assertEqual(c.get_checkin_compliance(),
                         ClientCompliance.STATE_UNKNOWN)
        self.assertEqual(c.get_disk_compliance(),
                         ClientCompliance.STATE_UNKNOWN)
        self.assertEqual(c.get_compliance_state(),
                         ClientCompliance.STATE_UNKNOWN)
        self.assertEqual(c.get_patch_compliance(),
                         ClientCompliance.STATE_UNKNOWN)

    def test_compliance_patch_status_ok(self):
        c = ClientCompliance(None, None, None, 0, 0)
        self.assertEqual(c.get_patch_compliance(),
                         ClientCompliance.STATE_OK)

    def test_compliance_patch_status_failure(self):
        # 1 critical package, no other packages
        c = ClientCompliance(None, None, None, 1, 0)
        self.assertEqual(c.get_patch_compliance(),
                         ClientCompliance.STATE_ERROR)
        # 0 critical packages, 1 other package
        c = ClientCompliance(None, None, None, 0, 1)  
        self.assertEqual(c.get_patch_compliance(),
                         ClientCompliance.STATE_WARNING)

    def test_compliance_checkin_less_than_7_is_ok(self):
        c = ClientCompliance(None, None, 6)
        self.assertEqual(c.get_checkin_compliance(), ClientCompliance.STATE_OK)
        c.last_checkin_days = 7
        self.assertEqual(c.get_checkin_compliance(),
                         ClientCompliance.STATE_WARNING)

    def test_compliance_checkin_less_than_14_is_warning(self):
        c = ClientCompliance(None, None, 13)
        self.assertEqual(c.get_checkin_compliance(),
                         ClientCompliance.STATE_WARNING)
        c.last_checkin_days = 14
        self.assertEqual(c.get_checkin_compliance(),
                         ClientCompliance.STATE_ERROR)

    def test_compliance_disk_space_free_more_than_20pcnt(self):
        # < 20% left is warning
        # < 5% left is critical, or less than 5g
        c = ClientCompliance(100, 25, 3)
        self.assertEqual(c.get_disk_compliance(), ClientCompliance.STATE_OK)
        c = ClientCompliance(100, 20, 3)
        self.assertEqual(c.get_disk_compliance(), ClientCompliance.STATE_OK)

    def test_compliance_disk_space_free_between_5_and_20pcnt(self):
        c = ClientCompliance(100, 5, 3)
        self.assertEqual(c.get_disk_compliance(),
                         ClientCompliance.STATE_WARNING)
        c = ClientCompliance(100, 19, 3)
        self.assertEqual(c.get_disk_compliance(),
                         ClientCompliance.STATE_WARNING)

    def test_compliance_disk_space_free_less_than_5pcnt(self):
        c = ClientCompliance(100, 4.9, 3)
        self.assertEqual(c.get_disk_compliance(), ClientCompliance.STATE_ERROR)

    def test_compliance_value_is_max_of_either(self):
        # enforce that it's max() of either value
        c1 = ClientCompliance(100, 5, 3)  # WARNING for DISK
        c2 = ClientCompliance(100, 40, 500)  # ERROR for CHECKIN DAYS
        self.assertTrue(c2.get_checkin_compliance() > c1.get_disk_compliance())
        self.assertEqual(c2.get_compliance_state(),
                         ClientCompliance.STATE_ERROR)


class TestConfiguration(unittest.TestCase):
    def setUp(self):
        self.cfg = ExtraMetricsConfiguration()
        self.config_text = '''
            [extra_metrics]
            fw_server_hostname = abc
            fw_server_api_key = def
            '''
        buf = io.StringIO(self.config_text)
        self.cfg.read_configuration(buf)

    def test_configuration_defaults(self):
        self.assertEqual("abc", self.cfg.get_fw_api_server())
        self.assertEqual("def", self.cfg.get_fw_api_key())

    def test_configuration_changes_are_written(self):
        self.cfg.set_fw_api_key('hello world')
        self.cfg.set_polling_delay_seconds(42)
        buf = io.StringIO()
        self.cfg.write_configuration(buf)
        value = buf.getvalue()
        self.assertTrue("hello world" in value)
        self.assertTrue("42" in value)
        self.assertEqual(self.cfg.get_polling_delay_seconds(), "42")

def get_unparsable_version():
    return "13.3.3"

def get_incorrect_version():
    return "fwxserver 13.3.3"

def get_correct_version():
    return "fwxserver 14.0.0"

class TestRuntimeChecks(unittest.TestCase):
    @mock.patch('extra_metrics.scripts.get_current_fw_version', get_unparsable_version)
    def test_validation_fails_due_to_bad_parsing(self):
        with self.assertRaises(ValidationExceptionCannotParseFileWaveVersion):
            validate_current_fw_version()

    @mock.patch('extra_metrics.scripts.get_current_fw_version', get_correct_version)
    def test_validation_succeeds(self):
        validate_current_fw_version()

    @mock.patch('extra_metrics.scripts.get_current_fw_version', get_incorrect_version)
    def test_validation_finds_incorrect_version(self):
        with self.assertRaises(ValidationExceptionWrongFileWaveVersion):
            validate_current_fw_version()

    def test_can_run_supervisord_provisioning(self):
        provision_supervisord_runtime()

if __name__ == "__main__":
    unittest.main()




query_client_info = '''{
    "criteria": {
        "expressions": [
            {
                "column": "filewave_id",
                "component": "Client",
                "operator": "!=",
                "qualifier": null
            },
            {
                "column": "archived",
                "component": "Client",
                "operator": "=",
                "qualifier": null
            },
            {
                "column": "filewave_client_version",
                "component": "DesktopClient",
                "operator": "!=",
                "qualifier": null
            }
        ],
        "logic": "all"
    },
    "fields": [
        {
            "column": "device_name",
            "component": "Client"
        },
        {
            "column": "filewave_client_locked",
            "component": "Client"
        },
        {
            "column": "free_disk_space",
            "component": "Client"
        },
        {
            "column": "device_id",
            "component": "Client"
        },
        {
            "column": "is_tracking_enabled",
            "component": "Client"
        },
        {
            "column": "location",
            "component": "Client"
        },
        {
            "column": "serial_number",
            "component": "Client"
        },
        {
            "column": "filewave_client_version",
            "component": "DesktopClient"
        },
        {
            "column": "management_mode",
            "component": "Client"
        },
        {
            "column": "filewave_client_name",
            "component": "Client"
        },
        {
            "column": "filewave_id",
            "component": "Client"
        },
        {
            "column": "version",
            "component": "OperatingSystem"
        },
        {
            "column": "enrollment_state",
            "component": "Client"
        },
        {
            "column": "name",
            "component": "OperatingSystem"
        },
        {
            "column": "edition",
            "component": "OperatingSystem"
        },
        {
            "column": "build",
            "component": "OperatingSystem"
        },
        {
            "column": "type",
            "component": "OperatingSystem"
        },
        {
            "column": "last_check_in",
            "component": "Client"
        },
        {
            "column": "filewave_model_number",
            "component": "DesktopClient"
        },
        {
            "column": "device_manufacturer",
            "component": "DesktopClient"
        },
        {
            "column": "last_logged_in_username",
            "component": "Client"
        },
        {
            "column": "device_product_name",
            "component": "Client"
        },
        {
            "column": "current_upstream_host",
            "component": "Client"
        },
        {
            "column": "current_upstream_port",
            "component": "Client"
        },
        {
            "column": "total_disk_space",
            "component": "Client"
        }

    ],
    "main_component": "Client",
    "display_name": "extra metrics - client info"
}
'''

query_software_patches = '''{
    "favorite": true,
    "fields": [
        {
            "column": "filewave_id",
            "component": "Client"
        },
        {
            "column": "filewave_client_name",
            "component": "Client"
        },
        {
            "column": "name",
            "component": "Update"
        },
        {
            "column": "update_id",
            "component": "Update"
        },
        {
            "column": "version",
            "component": "Update"
        },
        {
            "column": "platform",
            "component": "Update"
        },
        {
            "column": "critical",
            "component": "Update"
        }
    ],
    "main_component": "Update",
    "display_name": "extra metrics - software patch"
}'''



import requests
from prometheus_client import Histogram
from .logs import logging
from .queries import query_client_info, query_software_patches, query_client_info
import json


http_request_time_taken = Histogram('extra_metrics_http_request_time_taken',
                                    'Response time (s)', ['method'])


http_request_time_taken_get_client_info = http_request_time_taken.labels(
    'get_client_info')
http_request_time_taken_get_software_patches = http_request_time_taken.labels(
    'get_software_patches')
http_request_time_taken_get_applications = http_request_time_taken.labels(
    'get_applications')
http_request_time_taken_get_software_updates_web = http_request_time_taken.labels(
    'get_software_updates_web')


# TODO: all these methods should return either s or json, but how is one supposed to know from the name?

class FWRestQuery:
    def __init__(self, hostname, api_key):
        self.hostname = hostname
        self.api_key = api_key
        super()

    def _fw_run_inv_query(self, additional=None):
        return 'https://' + self.hostname + ':20445/inv/api/v1/' + additional

    def _fw_run_web_query(self, additional):
        return 'https://' + self.hostname + '/api/' + additional

    def _auth_headers(self):
        return {'Authorization': self.api_key, 'Content-Type': 'application/json'}

    def get_definition_for_query_id_j(self, query_id):
        r = requests.get(self._fw_run_inv_query(f'query/{query_id}'),
                             headers=self._auth_headers())
        if r.status_code == 200:
            return json.loads(r.content)
        return None

    def get_results_for_query_id(self, query_id):
        return requests.get(self._fw_run_inv_query(f'query_result/{query_id}'),
                             headers=self._auth_headers())

    def find_group_with_name(self, group_name):
        # get the group, is it there?
        r = requests.get(self._fw_run_web_query(
            'reports/groups_tree'), headers=self._auth_headers())
        if r.status_code == 200:
            for item in r.json()["groups_hierarchy"]:
                if item["name"] == group_name:
                    return item
        return None

    def ensure_inventory_query_group_exists(self, group_name):
        existing_group = self.find_group_with_name(group_name)
        if existing_group is not None:
            return existing_group, False

        group_create_data = json.dumps({"name": group_name})
        (f"data payload is: {group_create_data}")
        requests.post(self._fw_run_web_query('reports/groups/'),
                        headers=self._auth_headers(),
                        data=group_create_data)

        return self.find_group_with_name(group_name), True

    def get_all_inventory_queries(self):
        r = requests.get(self._fw_run_inv_query('query/'),
            headers=self._auth_headers())
        if r.status_code == 200:
            return json.loads(r.content)
        return None

    def create_inventory_query(self, json_str):
        # just create only, don't validate if it exists....
        return requests.post(self._fw_run_inv_query('query/'),
                            headers=self._auth_headers(),
                            data=json_str)

    @http_request_time_taken_get_client_info.time()
    def get_client_info(self):
        return requests.post(self._fw_run_inv_query('query_result/'),
                             headers=self._auth_headers(),
                             data=query_client_info)

    @http_request_time_taken_get_software_patches.time()
    def get_software_patches(self):
        return requests.post(self._fw_run_inv_query('query_result/'),
                             headers=self._auth_headers(),
                             data=query_software_patches)

    @http_request_time_taken_get_software_updates_web.time()
    def get_software_updates_web_ui(self):
        return requests.get(self._fw_run_web_query('updates/ui/?limit=10000'),
                            headers=self._auth_headers())



import configparser
import os 
from extra_metrics.logs import logger

def read_config_helper(cfg):
    logger.info(f"loading the configuration from file {ExtraMetricsConfiguration.DEFAULT_CFG_FILE_LOCATION}")
    with open(ExtraMetricsConfiguration.DEFAULT_CFG_FILE_LOCATION, 'r') as f:
        cfg.read_configuration(f)
        return True
    return False

class ExtraMetricsConfiguration:
    DEFAULT_CFG_FILE_LOCATION = '/usr/local/etc/filewave/extra_metrics.ini'
    KEY_FW_SERVER_HOSTNAME = 'fw_server_hostname'
    KEY_FW_SERVER_API_KEY = 'fw_server_api_key'
    KEY_POLLING_DELAY = 'fw_query_polling_delay_seconds'

    def __init__(self):
        self.config = configparser.ConfigParser()
        self.config.add_section('extra_metrics')
        self.section = self.config['extra_metrics']
                
    def read_configuration(self, file_obj):
        self.config.read_file(file_obj)
        self.section = self.config['extra_metrics']

    def write_configuration(self, file_obj):
        self.config.write(file_obj)
    
    def _get_value(self, attr_name, default=None):
        return self.section.get(attr_name, default)

    def _set_value(self, attr_name, value):
        self.section[attr_name] = value

    def get_fw_api_server(self):
        return self._get_value(ExtraMetricsConfiguration.KEY_FW_SERVER_HOSTNAME)
    def set_fw_api_server(self, value):
        self._set_value(ExtraMetricsConfiguration.KEY_FW_SERVER_HOSTNAME, value)

    def get_fw_api_key(self):    
        return self._get_value(ExtraMetricsConfiguration.KEY_FW_SERVER_API_KEY)
    def set_fw_api_key(self, value):
        self._set_value(ExtraMetricsConfiguration.KEY_FW_SERVER_API_KEY, value)

    def get_polling_delay_seconds(self):    
        # default is 5 mins
        return self._get_value(ExtraMetricsConfiguration.KEY_POLLING_DELAY, 30)
    def set_polling_delay_seconds(self, value):
        self._set_value(ExtraMetricsConfiguration.KEY_POLLING_DELAY, str(value))



from prometheus_client import Gauge
import pandas as pd
from extra_metrics.logs import logger

software_updates_by_state = Gauge('extra_metrics_software_updates_by_state', 
    'buckets of all the software updates, according to the number of devices in each state',
    ["state"])

software_updates_by_device = Gauge('extra_metrics_software_updates_by_device', 
    'list of devices and the number of [critical] updates they need to have installed', 
    ["device_name", "device_id", "is_update_critical"])

software_updates_by_platform = Gauge('extra_metrics_software_updates_by_platform', 
    'list of platforms and the number of [critical] updates they have available', 
    ["platform_name", "is_update_critical"])


class SoftwarePatchStatus:
    def __init__(self, fw_query):
        self.fw_query = fw_query
        self.state_by_patch = {}

    def collect_patch_data_status(self):
        r = self.fw_query.get_software_updates_web_ui()
        j = r.json()

        assert j["next"] is None
        r = j["results"]

        self.state_by_patch = {}

        # goal: number of software updates, grouped by status:
        # status values:
        # NEW: count_unassigned
        # IN PROGRESS: count_assigned
        # FAILED: assigned_clients.error + assigned_clients.warning
        # DONE: assigned_clients.remaining == 0

        counters = {
            "assigned": 0,
            "completed": 0,
            "remaining": 0,
            "warning": 0,
            "error": 0
        }

        for item in r:
            self.state_by_patch[item['update_id']] = item
            
            acc = item["assigned_clients_count"]
            counters["assigned"] += acc["assigned"]
            counters["completed"] += acc["completed"]
            counters["remaining"] += acc["remaining"]
            counters["warning"] += acc["warning"]
            counters["error"] += acc["error"]

        software_updates_by_state.labels('Assigned').set(counters["assigned"])
        software_updates_by_state.labels('Completed').set(counters["completed"])
        software_updates_by_state.labels('Remaining').set(counters["remaining"])
        software_updates_by_state.labels('Warning').set(counters["warning"])
        software_updates_by_state.labels('Error').set(counters["error"])
        
    def collect_patch_data_per_device(self):
        r = self.fw_query.get_software_patches()
        j = r.json()

        df = pd.DataFrame(j["values"], columns=j["fields"])
        platform_mapping = {
            "0": "Apple",
            "1": "Microsoft"
        }

        ru = df.groupby(["Client_filewave_client_name", "Client_filewave_id", "Update_critical"], as_index=False)["Update_name"].count()
        for i in ru.to_numpy():
            client_name = i[0]
            client_id = int(i[1])
            is_crit = i[2]
            update_count = i[3]
            logger.info(f"patches, device: {client_name}/{client_id}, critical: {is_crit}, number of updates: {update_count}")
            software_updates_by_device.labels(client_name, client_id, is_crit).set(update_count)

        ru = df.groupby(["Update_platform", "Update_critical"], as_index=False)["Update_name"].count()
        for i in ru.to_numpy():
            platform_name = platform_mapping[i[0]]
            is_crit = i[1]
            update_count = i[2]
            software_updates_by_platform.labels(platform_name, is_crit).set(update_count)



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="filewave-extra-metrics", 
    version="1.0.7",
    author="John Clayton",
    author_email="johnc@filewave.com",
    description="An additional module that exposes s/ware patching and metrics information to the built in FileWave dashboard",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/johncclayton/useful_dashboard_info",
    packages=setuptools.find_packages(),
    include_package_data=True,
    entry_points = {
        'console_scripts': [
            'extra-metrics-config=extra_metrics.scripts:install_into_environment',
            'extra-metrics-run=extra_metrics.main:serve_and_process',
        ]
    },
    install_requires=[
        'requests',
        'prometheus_client',
        'pandas',
        'PyYAML',
        'click',
        'progressbar2',
        'timeloop'
    ],
    setup_requires=[
        'wheel'
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.7',
)